/* 
   sharesecret - ShareSecret joins and splits secrets securely.

   Copyright (C) 2000 Stefan Karrmann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

$Id: sharesecret.c,v 1.3 2000/02/23 21:19:58 sk Exp $
$Log: sharesecret.c,v $
Revision 1.3  2000/02/23 21:19:58  sk
First working Version

Revision 1.2  2000/02/19 22:29:07  sk
bersetzbar - mit Laufzeitfehlern

Revision 1.1  2000/02/13 23:13:29  sk
worked on commandline options


#include <termios.h>
#include <grp.h>
#include <pwd.h>
*/

#include <stdio.h>
#include <sys/types.h>
#include <getopt.h>
#include "sharesecret.h"
#include "checkopt.h"

#define MPZ_ASCII_BASE (36)

inline
void
doGenerate ()
{
  FILE * secretFile;
  mpz_t secret, modulus;
  mpz_t * x, * y;      

  if ( HAVE_OPT( MIN))
    {
      if (OPT_VALUE_MIN > OPT_VALUE_SPLIT)
	error (0, ERR_WARN,
	       _("You never can join more parts than you split up"));
      if (2 > OPT_VALUE_MIN)
	{
	  error (0, ERR_WARN,
		 _("Joining only one part, "
		   "implies that it is the secret already.\n"
		   "Ignoring this value completely"));
	  CLEAR_OPT( MIN);
	}
    }
  if ( HAVE_OPT( SECRET))
    {
      secretFile = fopen (OPT_ARG( SECRET), "r");
      if (NULL ==  secretFile)
	error (EXIT_FAILURE, 0, _("Cannot open secret file"));
    }
  else
    secretFile = stdin;

  readSecret (secret, secretFile);
  fclose (secretFile);

  if ( HAVE_OPT( MIN) && OPT_VALUE_MIN < OPT_VALUE_SPLIT)
    generate (& x, & y, modulus, secret, OPT_VALUE_MIN, OPT_VALUE_SPLIT);
  else /* switch to faster algorithm */
      xor_generate (& y, secret, OPT_VALUE_SPLIT);
  
  if ( HAVE_OPT( OUTPUT))
    {
      size_t i;
      FILE  * out;
      char * name;
      for (i = 0; i < OPT_VALUE_SPLIT; i++)
	{
	  asprintf (& name, "%s.%03d", OPT_ARG( OUTPUT), i);
	  out = fopen (name, "w");
	  free (name);
	  if (HAVE_OPT( MIN) && OPT_VALUE_MIN != OPT_VALUE_SPLIT)
	    {
	      writePartPair (x[i], y[i], modulus,
			     OPT_VALUE_MIN, OPT_VALUE_SPLIT, out);
	      mpz_clear (x[i]);
	    }
	  else
	    writePartSingle (y[i], OPT_VALUE_SPLIT, out);
	  fclose (out);
	}
      if (OPT_VALUE_MIN != OPT_VALUE_SPLIT)
	free (x);
      free (y);
    }
  else
    USAGE( EXIT_FAILURE);
}

inline
void
doSolve (int joiners, char ** filenames)
{				/* joiners= degree of polynomial */
  FILE * secretFile;
  mpz_t secret, modulus;
  mpz_t * x, * y;      
  FILE * in;
  int min, split, split2, i;
  mpz_t modulus2;

  /* precondition */
  if (joiners < 2)
    USAGE( 0);

  /* init */
  y = xmalloc (sizeof( mpz_t) * joiners);
  x = xmalloc (sizeof( mpz_t) * joiners);
  mpz_init (modulus);
  split = 0;

  /* work */
  for (i = joiners; i > 0; i--)
    {
      char* p = *(filenames ++);
      in = fopen (p, "r");
      readPart (x[i - 1], y[i - 1], modulus2, & min, & split2, in);
      /* Test if  part is acceptable */
      /* Todo: accept overdeterminated joins */
      if (min != joiners || (0 != split && split != split2) || 0 == split2)
	error (EXIT_FAILURE, 0, _("Joining parts are inconsistent"));
      if (0 == split)
	{
	  split = split2;
	  if (joiners > split)
	    error (EXIT_FAILURE, 0, _("Joining is inconsistent "
				      "(more parts than generated)"));
	}
      if (joiners < split)
	{
	  if (0 == mpz_sgn (modulus2))
	    error (EXIT_FAILURE, 0,
		   _("Modulus 0 is inconsistent "
		     "(doesn't generate a finite field)"));
	  if (0 == mpz_sgn (modulus))
	    mpz_set (modulus, modulus2);
	  else if (0 != mpz_cmp (modulus, modulus2))
	    error (EXIT_FAILURE, 0, _("You are trying to join secret "
				      "parts of different secrets"));
	  mpz_clear (modulus2);
	}
      fclose (in);
    }

  mpz_init (secret);
  if (min != split)
    {
      solve (secret, modulus, joiners, x, y);
      do
	{
	  joiners = joiners - 1;
	  mpz_clear (x[joiners]);
	  mpz_clear (y[joiners]);
	}
      while (joiners > 0);
    }
  else
    {
      xor_solve (secret, joiners, y);
      do
	{
	  joiners = joiners - 1;
	  mpz_clear (y[joiners]);
	}
      while (joiners > 0);
    }
  free (x); free (y);

  if ( HAVE_OPT( OUTPUT))
    {
      secretFile = fopen (OPT_ARG( OUTPUT), "w");
      if (NULL ==  secretFile)
	error (EXIT_FAILURE, 0, _("Cannot open secret file"));
    }
  else
    secretFile = stdout;
  writeSecret (secret, secretFile);
  fclose (secretFile);
  mpz_clear (secret);
}

int
main (int argc, char **argv)
{
  {
    int arg_ct = optionProcess( &sharesecretOptions, argc, argv );
    argc -= arg_ct;
    /*
    if ((argc < ARGC_MIN) || (argc > ARGC_MAX)) 
      {
	fprintf( stderr, "%s ERROR:  remaining args (%d) "
		 "out of range\n", myprogOptions.pzProgName,
		 argc );
	
	USAGE( EXIT_FAILURE );
      }
    */
    argv += arg_ct;
  }

  switch (WHICH_IDX_SPLIT)
    {
    case INDEX_OPT_SPLIT:
      doGenerate ();
      break;
    case INDEX_OPT_JOIN:
      doSolve (argc, argv);
      break;
    default:
      USAGE( EXIT_FAILURE );
      break;
    }
  exit (0);
}

